CHUNK_SIZE = 1984

HELPER_VERSIONS = {
    '11.00.00'
}

ME_VERSIONS = {
    '11.00.25.230'
}

PRODUCT_CODES = {
    '102',
    '51'
}

PRODUCT_TYPES = {
    '24'
}

CREATE_DIR_SUCCESS = 183 # 'b\xb7'

GET_UNK1_VALUES = {
    b'\x02\x00'
}

GET_UNK2_VALUES = {
    b'\x01\x60'
}

GET_UNK3_VALUES = {
    b'\x64\x00'
}